use core::panic;
use std::error::Error;

#[derive(Debug, PartialEq, Eq)]
pub struct Grid<T: Clone> {
    pub rows: Vec<Vec<T>>,
    pub width: u32,
}

impl<T: Clone> Grid<T> {
    pub fn new(grid: Vec<Vec<T>>) -> Result<Self, Box<dyn Error>> {
        let row_length = grid[0].len();
        for row in &grid {
            if row_length != row.len() {
                return Err(Box::from("Rows need to all be equal in length"));
            }
        }
        Ok(Grid {
            rows: grid,
            width: row_length as u32,
        })
    }

    pub fn rotate_90(&mut self) -> Self {
        let height = self.rows.len();
        let width = self.rows[0].len();

        let mut new_grid = Vec::with_capacity(width);

        for col_idx in 0..width {
            let mut new_row = Vec::with_capacity(height);
            for row_idx in 0..height {
                new_row.push(self.rows[row_idx][col_idx].clone());
            }
            new_grid.push(new_row);
        }
        self.rows = new_grid;
        self.width = height as u32;
        Grid {
            rows: self.rows.clone(),
            width: height as u32,
        }
    }

    pub fn get_diagonals_aigu(&self) -> Vec<Vec<T>> {
        let mut diagonals = Vec::new();
        // col_idx == starting column of this diagonal
        for col_idx in 0..self.width {
            diagonals.push(self.get_diagonal_aigu(col_idx, 0));
            if col_idx == self.width - 1 {
                // row_idx == starting row of this diagonal
                for row_idx in 1..self.rows.len() as u32 {
                    diagonals.push(self.get_diagonal_aigu(col_idx, row_idx));
                }
            }
        }
        diagonals
    }

    fn get_diagonal_aigu(&self, starting_col: u32, starting_row: u32) -> Vec<T> {
        if starting_col != self.width - 1 && starting_row != 0 {
            panic!("Start of diagonal needs to be valid. Start at top row or rightmost column");
        }
        // col_idx == current col_idx of this diagonal
        let mut col_idx = starting_col as usize;
        // row_idx == current row_idx of this diagonal
        let mut row_idx = starting_row as usize;

        let mut diagonal = Vec::new();

        while let Some(row) = self.rows.get(row_idx) {
            if let Some(element) = row.get(col_idx) {
                diagonal.push(element.clone());
                if let Some(sub) = col_idx.checked_sub(1) {
                    col_idx = sub;
                } else {
                    break;
                }
                row_idx += 1;
            } else {
                break;
            }
        }

        diagonal.reverse();
        diagonal
    }

    pub fn get_diagonals_grave(&self) -> Vec<Vec<T>> {
        let mut diagonals = Vec::new();
        // col_idx == starting column of this diagonal
        for col_idx in (0..self.width).rev() {
            diagonals.push(self.get_diagonal_grave(col_idx, 0));
            if col_idx == 0 {
                // row_idx == starting row of this diagonal
                for row_idx in 1..self.rows.len() as u32 {
                    diagonals.push(self.get_diagonal_grave(col_idx, row_idx));
                }
            }
        }
        diagonals
    }

    fn get_diagonal_grave(&self, starting_col: u32, starting_row: u32) -> Vec<T> {
        if starting_col != 0 && starting_row != 0 {
            panic!("Start of diagonal needs to be valid. Start at top row or leftmost column");
        }
        // col_idx == current col_idx of this diagonal
        let mut col_idx = starting_col as usize;
        // row_idx == current row_idx of this diagonal
        let mut row_idx = starting_row as usize;

        let mut diagonal = Vec::new();

        while let Some(row) = self.rows.get(row_idx) {
            if let Some(element) = row.get(col_idx) {
                diagonal.push(element.clone());
                col_idx += 1;
                row_idx += 1;
            } else {
                break;
            }
        }

        diagonal
    }
}

#[cfg(test)]
mod test {
    use super::Grid;

    #[test]
    fn diag_aigu_4x4() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c", "d"],
            vec!["e", "f", "g", "h"],
            vec!["i", "j", "k", "l"],
            vec!["m", "n", "o", "p"],
        ])
        .unwrap();
        let res = vec![
            vec!["a"],
            vec!["e", "b"],
            vec!["i", "f", "c"],
            vec!["m", "j", "g", "d"],
            vec!["n", "k", "h"],
            vec!["o", "l"],
            vec!["p"],
        ];
        assert_eq!(res, grid.get_diagonals_aigu());
    }

    #[test]
    fn diag_aigu_4x3() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c", "d"],
            vec!["e", "f", "g", "h"],
            vec!["i", "j", "k", "l"],
        ])
        .unwrap();
        let res = vec![
            vec!["a"],
            vec!["e", "b"],
            vec!["i", "f", "c"],
            vec!["j", "g", "d"],
            vec!["k", "h"],
            vec!["l"],
        ];
        assert_eq!(res, grid.get_diagonals_aigu());
    }

    #[test]
    fn diag_aigu_3x4() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c"],
            vec!["d", "e", "f"],
            vec!["g", "h", "i"],
            vec!["j", "k", "l"],
        ])
        .unwrap();
        let res = vec![
            vec!["a"],
            vec!["d", "b"],
            vec!["g", "e", "c"],
            vec!["j", "h", "f"],
            vec!["k", "i"],
            vec!["l"],
        ];
        assert_eq!(res, grid.get_diagonals_aigu());
    }

    #[test]
    fn diag_grave_4x4() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c", "d"],
            vec!["e", "f", "g", "h"],
            vec!["i", "j", "k", "l"],
            vec!["m", "n", "o", "p"],
        ])
        .unwrap();
        let res = vec![
            vec!["d"],
            vec!["c", "h"],
            vec!["b", "g", "l"],
            vec!["a", "f", "k", "p"],
            vec!["e", "j", "o"],
            vec!["i", "n"],
            vec!["m"],
        ];
        assert_eq!(res, grid.get_diagonals_grave());
    }

    #[test]
    fn diag_grave_4x3() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c", "d"],
            vec!["e", "f", "g", "h"],
            vec!["i", "j", "k", "l"],
        ])
        .unwrap();
        let res = vec![
            vec!["d"],
            vec!["c", "h"],
            vec!["b", "g", "l"],
            vec!["a", "f", "k"],
            vec!["e", "j"],
            vec!["i"],
        ];
        assert_eq!(res, grid.get_diagonals_grave());
    }

    #[test]
    fn diag_grave_3x4() {
        let grid = Grid::new(vec![
            vec!["a", "b", "c"],
            vec!["d", "e", "f"],
            vec!["g", "h", "i"],
            vec!["j", "k", "l"],
        ])
        .unwrap();
        let res = vec![
            vec!["c"],
            vec!["b", "f"],
            vec!["a", "e", "i"],
            vec!["d", "h", "l"],
            vec!["g", "k"],
            vec!["j"],
        ];
        assert_eq!(res, grid.get_diagonals_grave());
    }

    #[test]
    fn rotate() {
        let mut grid = Grid::new(vec![vec!["a", "b"], vec!["c", "d"]]).unwrap();
        let res = Grid::new(vec![vec!["a", "c"], vec!["b", "d"]]).unwrap();
        assert_eq!(res, grid.rotate_90())
    }
}
